/*
 * cmd.c - contains CLI commands for status and config mgmt.
 *
 * Copyright (C) 2024 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <stdbool.h>
#include <errno.h>

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>

#include "debug.h"
#include "util.h"
#include "timer.h"
#include "csmngr.h"

static int csmngr_handle_command(struct csmngr_private *priv, int cli,
				char *cmdname, char *args)
{
	if (!strcmp(cmdname, "status")) {
		struct blob_buf bb = {0};
		char *jresp;

		blob_buf_init(&bb, 0);
		csmngr_status(priv, &bb);

		jresp = blobmsg_format_json(bb.head, true);
		if (!jresp) {
			blob_buf_free(&bb);
			return -1;
		}

		dprintf(cli, "%s\n", jresp);
		blob_buf_free(&bb);
		free(jresp);
	} else if (!strcmp(cmdname, "history")) {
		struct blob_buf bb = {0};
		char *jresp;

		blob_buf_init(&bb, 0);
		csmngr_dump_history(priv, &bb);

		jresp = blobmsg_format_json(bb.head, true);
		if (!jresp) {
			blob_buf_free(&bb);
			return -1;
		}

		dprintf(cli, "%s\n", jresp);
		blob_buf_free(&bb);
		free(jresp);
	} else if (!strcmp(cmdname, "add_radio")) {
		struct blob_buf bo = {0};
		struct blob_buf bi = {0};
		char *jresp;

		blob_buf_init(&bo, 0);
		blob_buf_init(&bi, 0);
		if (!blobmsg_add_json_from_string(&bi, args)) {
			dbg("%s: blobmsg from json-string error\n", __func__);
			blob_buf_free(&bi);
			return -1;
		}

		csmngr_add_radio(priv, bi.head, &bo);

		jresp = blobmsg_format_json(bo.head, true);
		if (!jresp) {
			blob_buf_free(&bo);
			blob_buf_free(&bi);
			return -1;
		}

		dprintf(cli, "%s\n", jresp);
		blob_buf_free(&bi);
		blob_buf_free(&bo);
		free(jresp);
	} else if (!strcmp(cmdname, "del_radio")) {
		struct blob_buf bo = {0};
		struct blob_buf bi = {0};
		char *jresp;

		blob_buf_init(&bo, 0);
		blob_buf_init(&bi, 0);
		if (args && strlen(args) && !blobmsg_add_json_from_string(&bi, args)) {
			dbg("%s: blobmsg from json-string error\n", __func__);
			blob_buf_free(&bi);
			return -1;
		}

		csmngr_del_radio(priv, bi.head, &bo);

		jresp = blobmsg_format_json(bo.head, true);
		if (!jresp) {
			blob_buf_free(&bo);
			blob_buf_free(&bi);
			return -1;
		}

		dprintf(cli, "%s\n", jresp);
		blob_buf_free(&bi);
		blob_buf_free(&bo);
		free(jresp);
	} else if (!strcmp(cmdname, "scan")) {
		struct blob_buf bo = {0};
		struct blob_buf bi = {0};
		char *jresp;

		blob_buf_init(&bo, 0);
		blob_buf_init(&bi, 0);
		if (args && strlen(args) && !blobmsg_add_json_from_string(&bi, args)) {
			dbg("%s: blobmsg from json-string error\n", __func__);
			blob_buf_free(&bi);
			return -1;
		}

		//TODO
		//csmngr_cmd_scan(priv, bi.head, &bo);

		jresp = blobmsg_format_json(bo.head, true);
		if (!jresp) {
			blob_buf_free(&bo);
			blob_buf_free(&bi);
			return -1;
		}

		dprintf(cli, "%s\n", jresp);
		blob_buf_free(&bi);
		blob_buf_free(&bo);
		free(jresp);
	} else {
		dprintf(cli, "Unknown command: %s\n", cmdname);
	}

	return 0;
}

static void csmngr_process_cmdline(struct csmngr_private *priv, int cli, char *buffer)
{
	char *cmdline, *cmdname, *args;
	char *line, *ptr, *saveptr;
	struct json_object *jmsg;
	int i = 0;


	ptr = buffer;
	line = strtok_r(ptr, "\n", &saveptr);
	if (!line)
		return;

	ptr = saveptr;
	if (strlen(line) == 0)
		return;

	cmdline = strdup(line);
	cmdname = strtok_r(cmdline, " ", &args);
	if (!cmdname) {
		dbg("%s: No command name\n", __func__);
		goto out;
	}

	dbg("%s: cli command: %s\n", __func__, cmdname);
	if (args && strlen(args)) {
		dbg("%s:        args: %s (len = %zu)\n", __func__, args, strlen(args));
		while (i <= strlen(args) && args[i++] == ' ');
		if (i <= strlen(args)) {
			/* json validate the passed args */
			jmsg = json_tokener_parse(args);
			if (!jmsg) {
				dbg("%s: Invalid json cmdline\n", __func__);
				goto out;
			}

			if (!json_object_is_type(jmsg, json_type_object)) {
				json_object_put(jmsg);
				goto out;
			}
		}
	}

	csmngr_handle_command(priv, cli, cmdname, args);
out:
	free(cmdline);
}

static void csmngr_recv_cmd(struct uloop_fd *fd, unsigned int events)
{
	struct csmngr_private *priv =
				container_of(fd, struct csmngr_private, cmdloop);

	for (;;) {
		struct sockaddr_un cliaddr = {0};
		char buffer[1024] = {0};
		socklen_t len = sizeof(struct sockaddr_un);
		int cli;
		int n;

		cli = accept(priv->cmdsock, (struct sockaddr *)&cliaddr, &len);
		if (cli < 0) {
			//dbg("%s: accept() %s\n", __func__, strerror(errno));
			return;
		}

		n = read(cli, buffer, sizeof(buffer));
		if (n == -1) {
			if (errno == EAGAIN || errno == EWOULDBLOCK)
				return;

			break;
		}

		if (n == 0)
			return;

		csmngr_process_cmdline(priv, cli, buffer);
		close(cli);
	}

	if (!priv->cmdloop.registered) {
		dbg("%s: uloop re-registering for cli commands\n", __func__);
		uloop_fd_delete(&priv->cmdloop);
		uloop_fd_add(&priv->cmdloop, ULOOP_READ /* | ULOOP_EDGE_TRIGGER */);
	}
}

int csmngr_start_cmdserver(struct csmngr_private *priv, struct csmngr_useropts *u)
{
	char cmdpath[96] = {0};
	struct sockaddr_un sa;
	int reuse = 1;
	int ret;
	int s;

	memset(&sa, 0, sizeof(sa));
	sa.sun_family = AF_UNIX;
	snprintf(cmdpath, sizeof(cmdpath), "%s.%ld", u->cmdpath, (long)getpid());
	strncpy(sa.sun_path, cmdpath, strlen(cmdpath));
	s = socket(AF_UNIX, SOCK_STREAM, 0);
	if (s < 0) {
		warn("%s: socket() %s\n", __func__, strerror(errno));
		return -1;
	}

	unlink(u->cmdpath);
	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(reuse)) != 0) {
		warn("%s: setsockopt() %s\n", __func__, strerror(errno));
		close(s);
		return -1;
	}

	if (bind(s, (struct sockaddr *)&sa, sizeof(sa)) != 0) {
		warn("%s: bind() %s\n", __func__, strerror(errno));
		close(s);
		return -1;
	}

	listen(s, 1);
	priv->cmdpath = strdup(cmdpath);
	priv->cmdsock = s;
	priv->cmdloop.fd = s;
	priv->cmdloop.cb = csmngr_recv_cmd;
	ret = uloop_fd_add(&priv->cmdloop, ULOOP_READ);
	if (ret) {
		close(s);
		priv->cmdsock = -1;
		warn("%s: uloop_fd_add() %s\n", __func__, strerror(errno));
		return -1;
	}

	dbg("%s: listen for commands on '%s'\n", __func__, u->cmdpath);
	return 0;
}

void csmngr_stop_cmdserver(struct csmngr_private *priv)
{
	if (priv && priv->cmdsock > 0) {
		uloop_fd_delete(&priv->cmdloop);
		priv->cmdloop.fd = -1;
		close(priv->cmdsock);
		unlink(priv->cmdpath);
	}
}
