/*
 * cmdresp.c - implement command handlers.
 *
 * Copyright (C) 2024 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if_arp.h>
#include <net/if.h>
#include <arpa/inet.h>

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/utils.h>

#include <easy/easy.h>

#include "debug.h"
#include "util.h"
#include "timer.h"
#include "csmngr.h"


static uint32_t bw_value(enum wifi_bw bw)
{
	switch (bw) {
	case BW20: return 20;
	case BW40: return 40;
	case BW80: return 80;
	case BW160: return 160;
	case BW320: return 320;
	case BW8080: return 8080;
	case BW_AUTO:
	case BW_UNKNOWN:
	default:
		return 0;
	}
}

static char *wifi_band_str(enum wifi_band band)
{
	switch (band) {
	case BAND_2:
		return "2g";
	case BAND_5:
		return "5g";
	case BAND_6:
		return "6g";
	default:
		break;
	}

	return "unknown";
}

static char *csmode_str(enum csmngr_mode mode)
{
	switch (mode) {
	case CS_MODE_DISABLED:
		return "disabled";
	case CS_MODE_ENABLED:
		return "enabled+acs";
	case CS_MODE_MONITOR:
		return "enabled";
	case CS_MODE_DRY_RUN:
		return "debug";
	default:
		break;
	}

	return "";
}

static char *cs_state_str(enum csmngr_state s)
{
	switch (s) {
	case CS_STATE_IDLE:
		return "idle";
	case CS_STATE_SCANNING:
		return "scanning";
	case CS_STATE_PROCESS:
		return "processing";
	case CS_STATE_CAC:
		return "cac";
	default:
		break;
	}

	return "";
}

static char *dfs_state_str(enum dfs_state state)
{
	switch (state) {
	case WIFI_DFS_STATE_USABLE:
		return "usable";
	case WIFI_DFS_STATE_CAC:
		return "cac";
	case WIFI_DFS_STATE_UNAVAILABLE:
		return "unavailable";
	case WIFI_DFS_STATE_AVAILABLE:
		return "available";
	default:
		break;
	}

	return "unknown";
}

int csmngr_status(struct csmngr_private *priv, void *out)
{
	struct blob_buf *bb = (struct blob_buf *)out;
	struct radio_interface *r = NULL;
	void *a, *t, *aa;

	a = blobmsg_open_array(bb, "radios");
	list_for_each_entry(r, &priv->radiolist, list) {
		char macstr[18] = {0};

		t = blobmsg_open_table(bb, "");
		hwaddr_ntoa(r->macaddr, macstr);
		blobmsg_add_string(bb, "macaddr", macstr);
		blobmsg_add_string(bb, "name", r->ifname);
		blobmsg_add_string(bb, "band", wifi_band_str(r->band));
		blobmsg_add_string(bb, "mode", csmode_str(r->mode));
		blobmsg_add_string(bb, "state", cs_state_str(r->state));
		blobmsg_add_u32(bb, "channel", r->info.channel);
		blobmsg_add_u32(bb, "bandwidth", bw_value(r->info.curr_bw));
		blobmsg_add_u64(bb, "channel_busy", r->info.diag.channel_busy / 1000);
		blobmsg_add_u64(bb, "channel_txtime", r->info.diag.tx_airtime / 1000);
		blobmsg_add_u64(bb, "channel_rxtime", r->info.diag.rx_airtime / 1000);
		blobmsg_add_u64(bb, "channel_obsstime", r->info.diag.obss_airtime / 1000);
		blobmsg_add_u64(bb, "channel_activetime", r->info.diag.cca_time / 1000);
		aa = blobmsg_open_array(bb, "channels");
		for (int j = 0; j < r->num_supp_channels; j++) {
			blobmsg_add_u32(bb, "", r->supp_channels[j]);
		}
		blobmsg_close_array(bb, aa);
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, a);

	return 0;
}

int csmngr_results(struct csmngr_private *priv, void *out)
{
	struct blob_buf *bb = (struct blob_buf *)out;
	struct radio_interface *r = NULL;
	void *a, *t, *aa, *tt;

	a = blobmsg_open_array(bb, "radios");
	list_for_each_entry(r, &priv->radiolist, list) {
		char macstr[18] = {0};

		t = blobmsg_open_table(bb, "");
		hwaddr_ntoa(r->macaddr, macstr);
		blobmsg_add_string(bb, "name", r->ifname);
		blobmsg_add_string(bb, "band", wifi_band_str(r->band));
		aa = blobmsg_open_array(bb, "verdict");
		for (int j = 0; j < NUM_BANDWIDTHS; j++) {
			if (r->verdict[j].channel == 0)
				continue;

			tt = blobmsg_open_table(bb, "");
			blobmsg_add_u32(bb, "channel", r->verdict[j].channel);
			blobmsg_add_u32(bb, "bandwidth", bw_value(r->verdict[j].bw));
			blobmsg_add_double(bb, "interference-factor", r->verdict[j].intfx);
			blobmsg_close_table(bb, tt);
		}
		blobmsg_close_array(bb, aa);
		blobmsg_close_table(bb, t);
	}
	blobmsg_close_array(bb, a);

	return 0;
}

int csmngr_dump_history(struct csmngr_private *priv, void *out)
{
	//struct blob_buf *bb = (struct blob_buf *)out;

	//TODO
	return 0;
}


/* add radio policy */
enum add_radio_attr {
	ADD_RADIO_ATTR_RADIO_NAME,	/* radio name */
	NUM_ATTRS_ADD_RADIO,
};

static const struct blobmsg_policy add_radio_policy[NUM_ATTRS_ADD_RADIO] = {
	[ADD_RADIO_ATTR_RADIO_NAME] = { .name = "radio", .type = BLOBMSG_TYPE_STRING },
};

int csmngr_add_radio(struct csmngr_private *priv, void *args, void *out)
{
	struct blob_attr *msg = (struct blob_attr *)args;
	struct blob_buf *bb = (struct blob_buf *)out;
	struct blob_attr *tb[NUM_ATTRS_ADD_RADIO];
	struct radio_interface *r;
	char radio[16] = {0};
	//int ret = 0;


	blobmsg_parse(add_radio_policy, NUM_ATTRS_ADD_RADIO, tb,
		      blob_data(msg), blob_len(msg));

	if (!tb[ADD_RADIO_ATTR_RADIO_NAME])
		return -EINVAL;

	strncpy(radio, blobmsg_data(tb[ADD_RADIO_ATTR_RADIO_NAME]), 16);
	radio[15] = '\0';

	r = csmngr_lookup_radio(priv, radio);
	if (r) {
		dbg("%s: %s is already known.\n", __func__, radio);
		return -EINVAL;
	}

	/* TODO
	r = csmngr_include_radio(priv, radio);
	if (!r) {
		dbg("%s: %s: error adding\n", __func__, radio);
		ret = 1;
	}
	*/

	/* reply with status */
	//blobmsg_add_string(bb, "status", ret == 0 ? "ok" : "fail");
	return 0;
}

/* del radio policy */
enum del_radio_attr {
	DEL_RADIO_ATTR_RADIO_NAME,	/* radio name */
	NUM_ATTRS_DEL_RADIO,
};

static const struct blobmsg_policy del_radio_policy[NUM_ATTRS_DEL_RADIO] = {
	[DEL_RADIO_ATTR_RADIO_NAME] = { .name = "radio", .type = BLOBMSG_TYPE_STRING },
};

int csmngr_del_radio(struct csmngr_private *priv, void *args, void *out)
{
	struct blob_attr *msg = (struct blob_attr *)args;
	struct blob_buf *bb = (struct blob_buf *)out;
	struct blob_attr *tb[NUM_ATTRS_DEL_RADIO];
	struct radio_interface *r;
	char radio[16] = {0};


	blobmsg_parse(del_radio_policy, NUM_ATTRS_DEL_RADIO, tb,
		      blob_data(msg), blob_len(msg));

	if (!tb[DEL_RADIO_ATTR_RADIO_NAME])
		return -EINVAL;

	strncpy(radio, blobmsg_data(tb[DEL_RADIO_ATTR_RADIO_NAME]), 16);
	radio[15] = '\0';

	r = csmngr_lookup_radio(priv, radio);
	if (!r) {
		dbg("%s: %s is not known.\n", __func__, radio);
		return -EINVAL;
	}

	/* TODO
	csmngr_exclude_radio(priv, radio);
	*/

	/* reply with status */
	blobmsg_add_string(bb, "status", "ok");
	return 0;
}

/* scan radio policy */
enum scan_attr {
	SCAN_ATTR_RADIO_NAME,		/* radio name */
	SCAN_ATTR_BAND,			/* band; 2, 5 or 6 (ignored when band provided) */
	SCAN_ATTR_CHANNEL,		/* 20MHz-channel number */
	SCAN_ATTR_FREQ,			/* Frequency in MHz (ignored when channel passed) */
	SCAN_ATTR_DURATION,		/* channel dwell-time in msecs */
	NUM_ATTRS_SCAN,
};

static const struct blobmsg_policy scan_policy[NUM_ATTRS_SCAN] = {
	[SCAN_ATTR_RADIO_NAME] = { .name = "radio", .type = BLOBMSG_TYPE_STRING },
	[SCAN_ATTR_BAND] = { .name = "band", .type = BLOBMSG_TYPE_INT32 },
	[SCAN_ATTR_CHANNEL] = { .name = "channel", .type = BLOBMSG_TYPE_INT32 },
	[SCAN_ATTR_FREQ] = { .name = "freq", .type = BLOBMSG_TYPE_INT32 },
	[SCAN_ATTR_DURATION] = { .name = "duration", .type = BLOBMSG_TYPE_INT32 },
};

int csmngr_scan(struct csmngr_private *priv, void *args, void *out)
{
	struct blob_attr *msg = (struct blob_attr *)args;
	struct blob_buf *bb = (struct blob_buf *)out;
	struct blob_attr *tb[NUM_ATTRS_SCAN];
	struct scan_param_ex sp = {0};
	struct radio_interface *r;
	char radio[16] = {0};
	uint32_t channel = 0;
	uint32_t freq = 0;
	uint8_t band = 0;
	int ret;


	blobmsg_parse(scan_policy, NUM_ATTRS_SCAN, tb,
		      blob_data(msg), blob_len(msg));

	if (!tb[SCAN_ATTR_RADIO_NAME] && !tb[SCAN_ATTR_BAND])
		return -EINVAL;

	if (!tb[SCAN_ATTR_CHANNEL] && !tb[SCAN_ATTR_FREQ])
		return -EINVAL;

	if (tb[SCAN_ATTR_RADIO_NAME]) {
		strncpy(radio, blobmsg_data(tb[SCAN_ATTR_RADIO_NAME]), 16);
		radio[15] = '\0';

		r = csmngr_lookup_radio(priv, radio);
	} else {
		band = (uint8_t)blobmsg_get_u32(tb[SCAN_ATTR_BAND]);
		if (band != 2 && band != 5 && band != 6)
			return -EINVAL;

		r = csmngr_lookup_radio_by_band(priv, band);
	}

	if (!r) {
		dbg("%s: %s is not known.\n", __func__, radio);
		return -EINVAL;
	}

	if (tb[SCAN_ATTR_FREQ]) {
		freq = blobmsg_get_u32(tb[SCAN_ATTR_FREQ]);
		//TODO: check if valid
	} else {
		channel = blobmsg_get_u32(tb[SCAN_ATTR_CHANNEL]);
		freq = wifi_channel_to_freq_ex(channel, r->band);
	}

	//TODO: duration

	sp.num_freq = 1;
	sp.freq[0] = freq;

	if (r->multiband)
		ret = wifi_scan_band_ex(r->phyname, r->band, &sp);
	else
		ret = wifi_scan_ex(r->phyname, &sp);

	/* reply with status */
	blobmsg_add_string(bb, "status", ret == 0 ? "ok" : "fail");
	return 0;
}

/* refresh policy */
enum refresh_attr {
	REFRESH_ATTR_RADIO_NAME,	/* radio name */
	REFRESH_ATTR_BAND,		/* band; 2, 5 or 6 (ignored when band provided) */
	NUM_ATTRS_REFRESH,
};

static const struct blobmsg_policy refresh_policy[NUM_ATTRS_REFRESH] = {
	[REFRESH_ATTR_RADIO_NAME] = { .name = "radio", .type = BLOBMSG_TYPE_STRING },
	[REFRESH_ATTR_BAND] = { .name = "band", .type = BLOBMSG_TYPE_INT32 },
};

int csmngr_refresh(struct csmngr_private *priv, void *args, void *out)
{
	struct blob_attr *msg = (struct blob_attr *)args;
	struct blob_buf *bb = (struct blob_buf *)out;
	struct blob_attr *tb[NUM_ATTRS_REFRESH];
	struct radio_interface *r;
	char radio[16] = {0};
	uint8_t band = 0;
	int ret = 0;


	blobmsg_parse(refresh_policy, NUM_ATTRS_REFRESH, tb,
		      blob_data(msg), blob_len(msg));

	if (!tb[REFRESH_ATTR_RADIO_NAME] && !tb[REFRESH_ATTR_BAND])
		return -EINVAL;

	if (tb[REFRESH_ATTR_RADIO_NAME]) {
		strncpy(radio, blobmsg_data(tb[REFRESH_ATTR_RADIO_NAME]), 16);
		radio[15] = '\0';

		r = csmngr_lookup_radio(priv, radio);
	} else {
		band = (uint8_t)blobmsg_get_u32(tb[REFRESH_ATTR_BAND]);
		if (band != 2 && band != 5 && band != 6)
			return -EINVAL;

		r = csmngr_lookup_radio_by_band(priv, band);
	}

	if (!r) {
		dbg("%s: %s is not known.\n", __func__, radio);
		return -EINVAL;
	}

	if (!r->running) {
		r->state = CS_STATE_IDLE;
		ret = csmngr_radio_run(r);
	}

	/* reply with status */
	blobmsg_add_string(bb, "status", ret == 0 ? "ok" : "fail");
	return 0;
}

/* dump survey policy */
enum dump_survey_attr {
	DUMP_SURVEY_ATTR_RADIO_NAME,		/* radio name */
	DUMP_SURVEY_ATTR_BAND,			/* band; 2, 5 or 6 (ignored when radio provided) */
	DUMP_SURVEY_ATTR_FREQ,			/* Frequency in MHz. */
	DUMP_SURVEY_ATTR_CHANNEL,		/* 20MHz-channel number (ignored when freq passed) */
	NUM_ATTRS_DUMP_SURVEY,
};

static const struct blobmsg_policy dump_survey_policy[NUM_ATTRS_DUMP_SURVEY] = {
	[DUMP_SURVEY_ATTR_RADIO_NAME] = { .name = "radio", .type = BLOBMSG_TYPE_STRING },
	[DUMP_SURVEY_ATTR_BAND] = { .name = "band", .type = BLOBMSG_TYPE_INT32 },
	[DUMP_SURVEY_ATTR_FREQ] = { .name = "freq", .type = BLOBMSG_TYPE_INT32 },
	[DUMP_SURVEY_ATTR_CHANNEL] = { .name = "channel", .type = BLOBMSG_TYPE_INT32 },
};

int csmngr_dump_survey(struct csmngr_private *priv, void *args, void *out)
{
	struct blob_attr *msg = (struct blob_attr *)args;
	struct blob_buf *bb = (struct blob_buf *)out;
	struct blob_attr *tb[NUM_ATTRS_DUMP_SURVEY];
	struct radio_interface *r;
	char radio[16] = {0};
	uint32_t channel = 0;
	uint32_t freq = 0;
	uint8_t band = 0;
	void *a, *tab;


	blobmsg_parse(dump_survey_policy, NUM_ATTRS_DUMP_SURVEY, tb,
		      blob_data(msg), blob_len(msg));

	if (!tb[DUMP_SURVEY_ATTR_RADIO_NAME] && !tb[DUMP_SURVEY_ATTR_BAND])
		return -EINVAL;

	if (tb[DUMP_SURVEY_ATTR_RADIO_NAME]) {
		strncpy(radio, blobmsg_data(tb[DUMP_SURVEY_ATTR_RADIO_NAME]), 16);
		radio[15] = '\0';

		r = csmngr_lookup_radio(priv, radio);
	} else {
		band = (uint8_t)blobmsg_get_u32(tb[DUMP_SURVEY_ATTR_BAND]);
		if (band != 2 && band != 5 && band != 6)
			return -EINVAL;

		r = csmngr_lookup_radio_by_band(priv, band);
	}

	if (!r) {
		dbg("%s: %s is not known.\n", __func__, radio);
		return -EINVAL;
	}

	if (tb[DUMP_SURVEY_ATTR_FREQ]) {
		freq = blobmsg_get_u32(tb[DUMP_SURVEY_ATTR_FREQ]);
		//TODO: check if valid
	}

	if (!freq && tb[DUMP_SURVEY_ATTR_CHANNEL]) {
		channel = blobmsg_get_u32(tb[DUMP_SURVEY_ATTR_CHANNEL]);
		freq = wifi_channel_to_freq_ex(channel, r->band);
	}

	dbg("Num-survey entries = %u\n", r->num_chinfo);
	a = blobmsg_open_array(bb, "survey_data");
	for (int i = 0; i < r->num_chinfo; i++) {
		struct chan_entry *e = &r->chinfo[i];

		if (!freq || freq == e->freq) {
			char tspbuf[64] = {0};
			struct tm *t = NULL;
			time_t now;

			tab = blobmsg_open_table(bb, "");
			blobmsg_add_u32(bb, "channel", e->channel);
			blobmsg_add_u32(bb, "freq", e->freq);

			t = gmtime(&r->tsp[i]);
			strftime(tspbuf, sizeof(tspbuf), "%Y-%m-%dT%H:%M:%SZ", t);
			blobmsg_add_string(bb, "timestamp", tspbuf);
			now = time(NULL);
			blobmsg_add_u64(bb, "age", (int)difftime(now, r->tsp[i]));

			blobmsg_add_u32(bb, "noise", e->noise);
			if (r->band == BAND_5) {
				blobmsg_add_u8(bb, "dfs", e->dfs);
				blobmsg_add_string(bb, "dfs_state", dfs_state_str(e->dfs_state));
				blobmsg_add_u32(bb, "cac_time", e->cac_time);
				if (e->dfs_state == WIFI_DFS_STATE_UNAVAILABLE)
					blobmsg_add_u32(bb, "nop_time", e->nop_time);
			}

			blobmsg_add_u64(bb, "busy", e->survey.channel_busy);
			blobmsg_add_u64(bb, "tx_time", e->survey.tx_airtime);
			blobmsg_add_u64(bb, "rx_time", e->survey.rx_airtime);
			blobmsg_add_u64(bb, "cca_time", e->survey.cca_time);
			blobmsg_close_table(bb, tab);
		}
	}
	blobmsg_close_array(bb, a);
	return 0;
}
