/*
 * config.c - csmngr config handling.
 *
 * Copyright (C) 2023 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/statfs.h>
#include <linux/magic.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>
#include <uci.h>
#include <uci_blob.h>

#include <easy/easy.h>

#include "debug.h"
#include "util.h"
#include "timer.h"
#include "csmngr.h"

struct wifimngr_device {
	char device[16];
	char phy[16];
	char band[16];
};

#define WIFI_DEV_MAX_NUM	4

struct wifimngr_device wifi_device[WIFI_DEV_MAX_NUM];
int wifi_device_num;

static int ieee80211_readint(const char *path)
{
	int fd;
	int rv = -1;
	char buffer[16];

	if ((fd = open(path, O_RDONLY)) > -1)
	{
		if (read(fd, buffer, sizeof(buffer)) > 0)
			rv = atoi(buffer);

		close(fd);
	}

	return rv;
}

static const char *ieee80211_phy_path_str(const char *phyname)
{
	static char path[PATH_MAX];
	const char *prefix = "/sys/devices/";
	int prefix_len = strlen(prefix);
	int buf_len, offset;
	struct dirent *e;
	char buf[512], *link;
	int phy_idx;
	int seq = 0;
	DIR *d;

	snprintf(buf, sizeof(buf), "/sys/class/ieee80211/%s/index", phyname);
	phy_idx = ieee80211_readint(buf);
	if (phy_idx < 0)
		return NULL;

	buf_len = snprintf(buf, sizeof(buf), "/sys/class/ieee80211/%s/device", phyname);
	link = realpath(buf, path);
	if (!link)
		return NULL;

	if (strncmp(link, prefix, prefix_len) != 0)
		return NULL;

	link += prefix_len;

	prefix = "platform/";
	prefix_len = strlen(prefix);
	if (!strncmp(link, prefix, prefix_len) && strstr(link, "/pci"))
		link += prefix_len;

	snprintf(buf + buf_len, sizeof(buf) - buf_len, "/ieee80211");
	d = opendir(buf);
	if (!d)
		return link;

	while ((e = readdir(d)) != NULL) {
		int cur_idx;

		snprintf(buf, sizeof(buf), "/sys/class/ieee80211/%s/index", e->d_name);
		cur_idx = ieee80211_readint(buf);
		if (cur_idx < 0)
			continue;

		if (cur_idx >= phy_idx)
			continue;

		seq++;
	}

	closedir(d);

	if (!seq)
		return link;

	offset = link - path + strlen(link);
	snprintf(path + offset, sizeof(path) - offset, "+%d", seq);

	return link;
}

static int find_phy_from_device_path(char *path, char *phy)
{
	bool found = false;
	struct dirent *p;
	DIR *d;


	d = opendir("/sys/class/ieee80211");
	if (!d)
		return -1;

	while ((p = readdir(d))) {
		const char *phypath;

		if (!strcmp(p->d_name, "."))
			continue;

		if (!strcmp(p->d_name, ".."))
			continue;

		phypath = ieee80211_phy_path_str(p->d_name);
		if (!phypath)
			continue;

		fprintf(stderr, "%s called |%s| vs |%s|\n", __func__, phypath, path);
		if (!strcmp(phypath, path)) {
			strncpy(phy, p->d_name, 15);
			found = true;
		}

		 if (found)
			 break;
	}

	closedir(d);
	return found ? 0 : -1;
}

static bool phy_dir_exist(char *phy)
{
	bool found = false;
	struct dirent *p;
	DIR *d;

	d = opendir("/sys/class/ieee80211");
	if (!d)
		return found;

	while ((p = readdir(d))) {
		if (!strcmp(p->d_name, "."))
			continue;

		if (!strcmp(p->d_name, ".."))
			continue;

		if (!strcmp(phy, p->d_name)) {
			found = true;
			break;
		}
	}

	closedir(d);
	return found;
}

static int uci_get_wifi_devices(char devlist[][16])
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;
	int n = 0;

	ctx = uci_alloc_context();
	if (!ctx)
		return -1;

	if (uci_load(ctx, "wireless", &pkg)) {
		uci_free_context(ctx);
		return -1;
	}

	/* flush older wifi_device */
	memset(&wifi_device, 0, WIFI_DEV_MAX_NUM * sizeof(struct wifimngr_device));
	wifi_device_num = 0;

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);
		if (!strcmp(s->type, "wifi-device")) {
			struct uci_element *x;
			struct uci_option *op;

			memset(devlist[n], 0, 16);
			strncpy(devlist[n], s->e.name, 15);

			strncpy(wifi_device[wifi_device_num].device, s->e.name, 15);
			strncpy(wifi_device[wifi_device_num].phy, s->e.name, 15);

			if (!phy_dir_exist(s->e.name)) {
				uci_foreach_element(&s->options, x) {
					op = uci_to_option(x);
					if (!strncmp(x->name, "path", 4) &&
					    op->type == UCI_TYPE_STRING) {
						char phy[16] = {0};
						int ret;

						ret = find_phy_from_device_path(op->v.string, phy);
						if (!ret) {
							memset(wifi_device[wifi_device_num].phy, 0, 15);
							strncpy(wifi_device[wifi_device_num].phy, phy, 15);
						}
					}
				}
			}

			/* Get band */
			uci_foreach_element(&s->options, x) {
				op = uci_to_option(x);
				if (!strncmp(x->name, "band", 4) &&
				    op->type == UCI_TYPE_STRING) {
					memset(wifi_device[wifi_device_num].band, 0, 15);
					strncpy(wifi_device[wifi_device_num].band, op->v.string, 15);
				}
			}

			wifi_device_num++;

			if (++n == WIFI_DEV_MAX_NUM)
				break;
		}
	}

	uci_free_context(ctx);
	return n;
}

static int wifimngr_get_wifi_devices(char devs[][16])
{
	return uci_get_wifi_devices(devs);
}

static char *wifi_radio_phyname(const char *radio)
{
	int i;

	if (!radio || radio[0] == '\0')
		return NULL;

	if (wifi_device_num == 0) {
		char radios[WIFI_DEV_MAX_NUM][16] = {0};
		int num_radios;

		num_radios = wifimngr_get_wifi_devices(radios);
		if (num_radios < 0)
			return NULL;
	}

	for (i = 0; i < WIFI_DEV_MAX_NUM; i++) {
		if (!strncmp(wifi_device[i].device, radio, 16))
			return wifi_device[i].phy;
	}

	return NULL;
}

uint8_t channels_2g[] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14 };

uint8_t channels_5g[] = { 36, 40, 44, 48, 52, 56, 60, 64,
			  100, 104, 108, 112, 116, 120, 124, 128,
			  132, 136, 140, 144, 149, 153, 157, 161,
			  165, 169, 173, 177 };

uint8_t channels_6g[] = { 1, 5, 9, 13, 17, 21, 25, 29,
			  33, 37, 41, 45, 49, 53, 57, 61,
			  65, 69, 73, 77, 81, 85, 89, 93,
			  97, 101, 105, 109, 113, 117, 121, 125,
			  129, 133, 137, 141, 145, 149, 153, 157,
			  161, 165, 169, 173, 177, 181, 185, 189,
			  193, 197, 201, 205, 209, 213, 217, 221,
			  225, 229, 233 };

void clear_pref_channels(struct radio_interface *r)
{
	memset(r->pref_channels, 0, sizeof(r->pref_channels));
	r->num_pref_channels = 0;
}

int fill_pref_channels(enum wifi_band band, const char *config_channels,
		       uint8_t *preflist, int *num)
{
	char *str = strdup(config_channels);
	char *s, *tmp;
	int i;

	*num = 0;
	foreach_token_r(s, str, tmp, ", ") {
		if (strstr(s, "-")) {
			int ch1 = 0, ch2 = 0;
			char *c, *tmp2;

			foreach_token_r(c, s, tmp2, "- ") {
				size_t chlen;
				uint8_t *chs;
				int ch;

				if (band == BAND_2) {
					chs = channels_2g;
					chlen = ARRAY_SIZE(channels_2g);
				} else if (band == BAND_5) {
					chs = channels_5g;
					chlen = ARRAY_SIZE(channels_5g);
				} else if (band == BAND_6) {
					chs = channels_6g;
					chlen = ARRAY_SIZE(channels_6g);
				} else
					return -1;

				ch = atoi(c);
				if (ch1 == 0)
					ch1 = ch;
				else if (ch2 == 0)
					ch2 = ch;
				else {
					return -1;
				}

				for (i = 0; i < chlen; i++) {
					if (ch == chs[i])
						break;
				}

				if (i == chlen)
					return -1;

				for (i = 0; i < chlen; i++) {
					if (chs[i] >= ch1 && chs[i] <= ch2)
						preflist[(*num)++] = chs[i];
				}
			}
		} else {
			size_t chlen;
			uint8_t *chs;
			int ch;

			if (band == BAND_2) {
				chs = channels_2g;
				chlen = ARRAY_SIZE(channels_2g);
			} else if (band == BAND_5) {
				chs = channels_5g;
				chlen = ARRAY_SIZE(channels_5g);
			} else if (band == BAND_6) {
				chs = channels_6g;
				chlen = ARRAY_SIZE(channels_6g);
			} else
				return -1;

			ch = atoi(s);
			for (i = 0; i < chlen; i++) {
				if (ch == chs[i]) {
					preflist[(*num)++] = chs[i];
					break;
				}
			}

			if (i == chlen)
				return -1;
		}
	}

	if (str)
		free(str);

	return 0;
}

static int csmngr_get_wifi_device_config(struct csmngr_private *priv, struct uci_section *s)
{
	enum {
		WIFI_DEVICE_ATTR_DISABLED,
		WIFI_DEVICE_ATTR_CHANNEL,
		WIFI_DEVICE_ATTR_CHANNELS,
		WIFI_DEVICE_ATTR_BAND,
		WIFI_DEVICE_ATTR_COUNTRY,
		NUM_WIFI_DEVICE_ATTRS,
	};
	const struct uci_parse_option wifi_device_attrs[NUM_WIFI_DEVICE_ATTRS] = {
		[WIFI_DEVICE_ATTR_DISABLED] = { .name = "disabled", .type = UCI_TYPE_STRING },
		[WIFI_DEVICE_ATTR_CHANNEL] = { .name = "channel", .type = UCI_TYPE_STRING },
		[WIFI_DEVICE_ATTR_CHANNELS] = { .name = "channels", .type = UCI_TYPE_STRING },
		[WIFI_DEVICE_ATTR_BAND] = { .name = "band", .type = UCI_TYPE_STRING },
		[WIFI_DEVICE_ATTR_COUNTRY] = { .name = "country", .type = UCI_TYPE_STRING },
	};

	struct uci_option *tb[NUM_WIFI_DEVICE_ATTRS];
	struct radio_interface *radio = NULL;
	char wifi_device_name[16] = {0};
	char country[4] = {0};
	bool disabled = false;
	int radio_band = 0;
	const char *chan = NULL;
	const char *chanlist = NULL;
	bool autochannel = false;


	uci_parse_section(s, wifi_device_attrs, NUM_WIFI_DEVICE_ATTRS, tb);

	if (!tb[WIFI_DEVICE_ATTR_CHANNEL]) {
		warn("wifi-device: 'channel' option is mandatory.\n");
		return -1;
	}

	chan = tb[WIFI_DEVICE_ATTR_CHANNEL]->v.string;
	if (!strncmp(chan, "auto", 4) || !strncmp(chan, "0", 1)) {
		autochannel = true;
	}

	dbg("wifi-device: update '%s' autochannel -> %d\n", s->e.name, autochannel);

	/* uci section name => wifi-device name */
	strncpy(wifi_device_name, s->e.name, strlen(s->e.name));

	if (tb[WIFI_DEVICE_ATTR_CHANNELS])
		chanlist = tb[WIFI_DEVICE_ATTR_CHANNELS]->v.string;

	if (tb[WIFI_DEVICE_ATTR_BAND]) {
		char *val = tb[WIFI_DEVICE_ATTR_BAND]->v.string;

		if (!strncmp(val, "2g", 2))
			radio_band = BAND_2;
		else if (!strncmp(val, "5g", 2))
			radio_band = BAND_5;
		else if (!strncmp(val, "6g", 2))
			radio_band = BAND_6;
		else {
			warn("config: 'band' invalid.\n");
			return -1;
		}
	}


	if (tb[WIFI_DEVICE_ATTR_DISABLED]) {
		char *val = tb[WIFI_DEVICE_ATTR_DISABLED]->v.string;

		disabled = atoi(val) == 1 ? true : false;
	}

	dbg("%s: %s disabled = %d ############\n", __func__, wifi_device_name, disabled);

	if (tb[WIFI_DEVICE_ATTR_COUNTRY]) {
		char *val = tb[WIFI_DEVICE_ATTR_COUNTRY]->v.string;

		if (strlen(val) > 0 && strlen(val) < 4) {
			strncpy(country, val, strlen(val));
		} else {
			warn("wifi-device: 'country' invalid.\n");
			return -1;
		}
	}

	radio = csmngr_lookup_radio(priv, wifi_device_name);
	if (!radio) {
		radio = csmngr_alloc_radio(priv, wifi_device_name);
		if (!radio)
			return -1;

		priv->num_radio++;
		list_add_tail(&radio->list, &priv->radiolist);
	}

#if 1	//XXX - improve
	memset(radio->phyname, 0, sizeof(radio->phyname));
	char *phyname = wifi_radio_phyname(wifi_device_name);
	if (phyname) {
		strncpy(radio->phyname, phyname, strlen(phyname));
		dbg("radio = %s, phy = %s\n", radio->ifname, radio->phyname);
	}
#endif

	radio->enabled = !disabled;
	radio->band = radio_band;
	radio->autochannel = autochannel;
	strncpy(radio->country, country, strlen(country));
	clear_pref_channels(radio);
	if (chanlist) {
		int ret;

		ret = fill_pref_channels(radio_band, chanlist, radio->pref_channels,
					 &radio->num_pref_channels);
		if (ret) {
			warn("%s: invalid channels in config.\n", radio->ifname);
			clear_pref_channels(radio);
		}
	}

	return 0;
}

static int csmngr_get_wifi_acs_config(struct csmngr_private *priv, struct uci_section *s)
{
	enum {
		RADIO_ATTR_ENABLED,
		RADIO_ATTR_NAME,
		RADIO_ATTR_BAND,
		RADIO_ATTR_MODE,
		RADIO_ATTR_REFRESH_INT,
		RADIO_ATTR_COLLECT_INT,
		RADIO_ATTR_CHANNELS,
		RADIO_ATTR_EXCLUDE_OPCLASS,
		RADIO_ATTR_EXCLUDE_CHANNELS,
		RADIO_ATTR_EXCLUDE_BANDWIDTH,
		RADIO_ATTR_EXCLUDE_DFS,
		RADIO_ATTR_EXCLUDE_NON_PSC,
		RADIO_ATTR_CHANNEL_BIAS,
		RADIO_ATTR_PUNCTURE_BITMAP,
		RADIO_ATTR_PUNCTURE_THRESHOLD,
		RADIO_ATTR_CAC_LIST,
		RADIO_ATTR_FIXED_CHANNEL,
		RADIO_ATTR_FIXED_BANDWIDTH,
		NUM_RADIO_ATTRS,
	};
	const struct uci_parse_option radio_attrs[NUM_RADIO_ATTRS] = {
		[RADIO_ATTR_ENABLED] = { .name = "enabled", .type = UCI_TYPE_STRING },
		[RADIO_ATTR_NAME] = { .name = "radio", .type = UCI_TYPE_STRING },
		[RADIO_ATTR_BAND] = { .name = "band", .type = UCI_TYPE_STRING },
		[RADIO_ATTR_MODE] = { .name = "mode", .type = UCI_TYPE_STRING },
		[RADIO_ATTR_REFRESH_INT] = { .name = "refresh_int", .type = UCI_TYPE_STRING },
		[RADIO_ATTR_COLLECT_INT] = { .name = "collect_int", .type = UCI_TYPE_STRING },
		[RADIO_ATTR_CHANNELS] = { .name = "channels", .type = UCI_TYPE_LIST },
		[RADIO_ATTR_EXCLUDE_OPCLASS] = { .name = "exclude_opclass", .type = UCI_TYPE_LIST },
		[RADIO_ATTR_EXCLUDE_CHANNELS] = { .name = "exclude_channels", .type = UCI_TYPE_LIST },
		[RADIO_ATTR_EXCLUDE_BANDWIDTH] = { .name = "exclude_bandwidths", .type = UCI_TYPE_LIST },
		[RADIO_ATTR_EXCLUDE_DFS] = { .name = "exclude_dfs", .type = UCI_TYPE_STRING },
		[RADIO_ATTR_EXCLUDE_NON_PSC] = { .name = "exclude_non_psc", .type = UCI_TYPE_STRING },
		[RADIO_ATTR_CHANNEL_BIAS] = { .name = "channel_bias", .type = UCI_TYPE_STRING },
		[RADIO_ATTR_PUNCTURE_BITMAP] = { .name = "puncture_bitmap", .type = UCI_TYPE_STRING },
		[RADIO_ATTR_PUNCTURE_THRESHOLD] = { .name = "puncture_threshold", .type = UCI_TYPE_STRING },
		[RADIO_ATTR_CAC_LIST] = { .name = "cac_list", .type = UCI_TYPE_LIST },
	};

	struct uci_option *tb[NUM_RADIO_ATTRS];
	struct radio_interface *radio = NULL;
	char radio_ifname[16] = {0};
	bool exclude_non_psc = true;
	bool exclude_dfs = false;
	const char *name = NULL;
	bool enabled = true;
	int refresh_int = 0;
	int collect_int = 0;
	int radio_band = 0;


	uci_parse_section(s, radio_attrs, NUM_RADIO_ATTRS, tb);

	if (!tb[RADIO_ATTR_NAME] && !tb[RADIO_ATTR_BAND]) {
		warn("config: either 'radio' or 'band' option is mandatory.\n");
		return -1;
	}

	if (tb[RADIO_ATTR_NAME]) {
		name = tb[RADIO_ATTR_NAME]->v.string;

		if (!strlen(name) || strlen(name) > 16) {
			dbg("config: invalid 'radio' name\n");
			return -1;
		}

		strncpy(radio_ifname, name, strlen(name));
	}

	if (tb[RADIO_ATTR_BAND]) {
		char *val = tb[RADIO_ATTR_BAND]->v.string;

		if (!strncmp(val, "2g", 2))
			radio_band = 2;
		else if (!strncmp(val, "5g", 2))
			radio_band = 5;
		else if (!strncmp(val, "6g", 2))
			radio_band = 6;
		else {
			warn("config: 'band' invalid.\n");
			return -1;
		}

		//TOOD: validate radio and band
	}

	if (tb[RADIO_ATTR_ENABLED]) {
		char *val = tb[RADIO_ATTR_ENABLED]->v.string;

		enabled = atoi(val) == 1 ? true : false;
	}

	if (tb[RADIO_ATTR_REFRESH_INT]) {
		char *val = tb[RADIO_ATTR_REFRESH_INT]->v.string;
		int ival = atoi(val);

		if (ival >= 0)
			refresh_int = ival;
		else
			dbg("config: 'refresh_int' invalid.\n");
	}

	if (tb[RADIO_ATTR_COLLECT_INT]) {
		char *val = tb[RADIO_ATTR_COLLECT_INT]->v.string;
		int ival = atoi(val);

		if (ival >= 0)
			collect_int = ival;
		else
			dbg("config: 'collect_int' invalid.\n");
	}

	if (tb[RADIO_ATTR_EXCLUDE_DFS]) {
		char *val = tb[RADIO_ATTR_EXCLUDE_DFS]->v.string;

		exclude_dfs = atoi(val) == 1 ? true : false;
	}

	if (tb[RADIO_ATTR_EXCLUDE_NON_PSC]) {
		char *val = tb[RADIO_ATTR_EXCLUDE_NON_PSC]->v.string;

		exclude_non_psc = atoi(val) == 1 ? true : false;
	}

	//TODO: other config

	radio = csmngr_lookup_radio(priv, radio_ifname);
	if (!radio) {
		radio = csmngr_alloc_radio(priv, radio_ifname);
		if (!radio)
			return -1;

		priv->num_radio++;
	}

	/* update acs config for radio */
	radio->enabled = enabled;
	radio->band = radio_band;
	radio->exclude_dfs = exclude_dfs;
	radio->exclude_non_psc= exclude_non_psc;
	radio->collect_int = collect_int;
	radio->refresh_int = refresh_int;

	return 0;
}

void csmngr_config_free(struct csmngr_private *priv)
{
	//Nothing
}

int csmngr_reconfig(struct csmngr_private *priv, const char *path, const char *file)
{
	struct uci_context *ctx;
	struct uci_package *pkg;
	struct uci_element *e;
	int ret = 0;

	ctx = uci_alloc_context();
	if (!ctx)
		return -1;

	if (path) {
		uci_set_confdir(ctx, path);
		fprintf(stderr, "config path: %s  file: %s\n", path, file);
	}

	if (uci_load(ctx, file, &pkg)) {
		uci_free_context(ctx);
		return -1;
	}

	uci_foreach_element(&pkg->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (!strcmp(s->type, "wifi-device"))
			ret |= csmngr_get_wifi_device_config(priv, s);

		if (!strcmp(s->type, "wifi-acs"))
			ret |= csmngr_get_wifi_acs_config(priv, s);
	}

	uci_free_context(ctx);
	return ret;
}

int csmngr_config_defaults(struct csmngr_private *priv)
{
	struct csmngr_config *cfg = &priv->cfg;

	cfg->enabled = true;
	cfg->history = true;
	cfg->history_persist = true;
	strncpy(cfg->history_file, DEFAULT_HISTORY_FILE, sizeof(cfg->history_file) - 1);
	cfg->history_ageout = DEFAULT_HISTORY_AGEOUT;
	cfg->history_maxnum = DEFAULT_HISTORY_MAX_NUM;
	cfg->history_maxsize = DEFAULT_HISTORY_MAX_SIZE;

	return 0;
}

static int csmngr_dump_radio_config(struct radio_interface *r)
{
	fprintf(stderr, "\tname              : %s\n", r->ifname);
	fprintf(stderr, "\tenabled           : %d\n", r->enabled);
	fprintf(stderr, "\tacs_mode          : %d\n", r->mode);
	fprintf(stderr, "\tband              : %d\n", r->band);
	fprintf(stderr, "\texclude_dfs       : %s\n", r->exclude_dfs ? "true" : "false");
	fprintf(stderr, "\texclude_non_psc   : %s\n", r->exclude_non_psc ? "true" : "false");
	if (r->num_pref_channels) {
		fprintf(stderr, "\tpreferred-channels: ");
		for (int i = 0; i < r->num_pref_channels; i++) {
			fprintf(stderr, "%u, ", r->pref_channels[i]);
		}
		fprintf(stderr, "\n");
	}
	return 0;
}

int csmngr_dump_config(struct csmngr_private *priv)
{
	struct csmngr_config *cfg = &priv->cfg;

	fprintf(stderr, "configuration\n");
	fprintf(stderr, "---\n");
	fprintf(stderr, " enabled           : %d\n", cfg->enabled);
	fprintf(stderr, " history           : %d\n", cfg->history);
	fprintf(stderr, " history persists  : %d\n", cfg->history_persist);
	fprintf(stderr, " history file      : %s\n", cfg->history_file);
	fprintf(stderr, " history maxnum    : %u\n", cfg->history_maxnum);
	fprintf(stderr, " history maxsize   : %u\n", cfg->history_maxsize);
	fprintf(stderr, " radio interfaces  :\n");

	if (priv->num_radio) {
		struct radio_interface *r;

		list_for_each_entry(r, &priv->radiolist, list) {
			csmngr_dump_radio_config(r);
		}
	}
	fprintf(stderr, "---\n");
	return 0;
}
