/*
 * csmngr.c - WiFi channel selection manager.
 *
 * Copyright (C) 2024 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if_arp.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <linux/if_packet.h>
#include <net/ethernet.h>

#include <netlink/route/link.h>

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>

#include <easy/easy.h>
#include <wifi.h>

#include "util.h"
#include "debug.h"
#include "timer.h"
#include "csmngr.h"


static int signal_pending;

static void csmngr_sighandler(int sig)
{
	signal_pending = sig;
}


static uint32_t bw_value(enum wifi_bw bw)
{
	switch (bw) {
	case BW20: return 20;
	case BW40: return 40;
	case BW80: return 80;
	case BW160: return 160;
	case BW320: return 320;
	case BW8080: return 8080;
	case BW_AUTO:
	case BW_UNKNOWN:
	default:
		return 0;
	}
}

static char *dfs_state_str(enum dfs_state state)
{
	switch (state) {
	case WIFI_DFS_STATE_USABLE:
		return "usable";
	case WIFI_DFS_STATE_CAC:
		return "cac";
	case WIFI_DFS_STATE_UNAVAILABLE:
		return "unavailable";
	case WIFI_DFS_STATE_AVAILABLE:
		return "available";
	default:
		break;
	}

	return "unknown";
}


int csmngr_dump_channel_preferences(int cli, struct radio_interface *r,
				    int num_opclass,
				    struct wifi_opclass *supp_opclass)
{
	void *a, *t, *aa, *tt, *aaa;
	struct blob_buf bb = {0};
	int i, j, k;
	char *jresp;

	blob_buf_init(&bb, 0);
	a = blobmsg_open_array(&bb, "pref_opclass");
	for (i = 0; i < num_opclass; i++) {
		t = blobmsg_open_table(&bb, "");
		blobmsg_add_u32(&bb, "opclass", supp_opclass[i].g_opclass);
		blobmsg_add_u32(&bb, "bandwidth", bw_value(supp_opclass[i].bw));
		blobmsg_add_u32(&bb, "txpower", supp_opclass[i].opchannel.txpower);
		aa = blobmsg_open_array(&bb, "channels");
		for (j = 0; j < supp_opclass[i].opchannel.num; j++) {
			tt = blobmsg_open_table(&bb, "");
			blobmsg_add_u32(&bb, "channel", supp_opclass[i].opchannel.ch[j].channel);
			blobmsg_add_u32(&bb, "score", supp_opclass[i].opchannel.ch[j].score);
			blobmsg_add_u32(&bb, "dfs", supp_opclass[i].opchannel.ch[j].dfs);
			if (supp_opclass[i].opchannel.ch[j].dfs) {
				blobmsg_add_string(&bb, "dfs_state", dfs_state_str(supp_opclass[i].opchannel.ch[j].dfs_state));
				blobmsg_add_u32(&bb, "cac_time", supp_opclass[i].opchannel.ch[j].cac_time);
				if (supp_opclass[i].opchannel.ch[j].dfs_state == WIFI_DFS_STATE_UNAVAILABLE)
					blobmsg_add_u32(&bb, "nop_time", supp_opclass[i].opchannel.ch[j].nop_time);
			}
			aaa = blobmsg_open_array(&bb, "ctrl_channels");
			for (k = 0; k < ARRAY_SIZE(supp_opclass[i].opchannel.ch[j].ctrl_channels); k++) {
				if (supp_opclass[i].opchannel.ch[j].ctrl_channels[k])
					blobmsg_add_u32(&bb, "", supp_opclass[i].opchannel.ch[j].ctrl_channels[k]);
			}
			blobmsg_close_array(&bb, aaa);
			blobmsg_close_table(&bb, tt);
		}
		blobmsg_close_array(&bb, aa);
		blobmsg_close_table(&bb, t);
	}
	blobmsg_close_array(&bb, a);

	jresp = blobmsg_format_json_indent(bb.head, true, 1);
	if (!jresp) {
		blob_buf_free(&bb);
		return -1;
	}

	dprintf(cli, "%s\n", jresp);
	blob_buf_free(&bb);
	free(jresp);

	return 0;
}

int csmngr_dump_channels_info(int cli, struct radio_interface *r, int num,
			      struct chan_entry *channels,
			      uint32_t filter_channel)
{
	struct blob_buf bb = {0};
	char *jresp;
	void *a, *t;

	blob_buf_init(&bb, 0);
	a = blobmsg_open_array(&bb, "channels");
	for (int i = 0; i < num; i++) {
		struct chan_entry *e = &channels[i];

		if (filter_channel && !(filter_channel == e->channel))
			continue;

		t = blobmsg_open_table(&bb, "");
		blobmsg_add_u32(&bb, "channel", e->channel);
		blobmsg_add_u32(&bb, "freq", e->freq);
		blobmsg_add_u32(&bb, "noise", e->noise);
		if (e->dfs) {
			blobmsg_add_u8(&bb, "radar", e->dfs);
			blobmsg_add_string(&bb, "dfs_state", dfs_state_str(e->dfs_state));
			blobmsg_add_u32(&bb, "cac_time", e->cac_time);
			if (e->dfs_state == WIFI_DFS_STATE_UNAVAILABLE)
				blobmsg_add_u32(&bb, "nop_time", e->nop_time);
		}

		blobmsg_add_u64(&bb, "busy_time", e->survey.channel_busy);
		blobmsg_add_u64(&bb, "tx_time", e->survey.tx_airtime);
		blobmsg_add_u64(&bb, "rx_time", e->survey.rx_airtime);
		blobmsg_add_u64(&bb, "obss_time", e->survey.obss_airtime);
		blobmsg_add_u64(&bb, "active_time", e->survey.cca_time);
		blobmsg_close_table(&bb, t);
	}
	blobmsg_close_array(&bb, a);

	jresp = blobmsg_format_json_indent(bb.head, true, 1);
	if (!jresp) {
		blob_buf_free(&bb);
		return -1;
	}

	dprintf(cli, "%s\n", jresp);
	blob_buf_free(&bb);
	free(jresp);
	return 0;
}

static void csmngr_radio_periodic_refresh(atimer_t *t)
{
	struct radio_interface *r = container_of(t, struct radio_interface, refreshtimer);

	if (!r->running) {
		csmngr_radio_run(r);
		timer_set(t, r->refresh_int * 1000);
	} else {
		/* a refresh is ongoing, defer by couple of hours */
		timer_set(t,  CSMNGR_DEFER_REFRESH_INT * 1000);
	}
}

int csmngr_radio_update_channels_info(struct radio_interface *r, int num, struct chan_entry *chinfo)
{
	bool changed = false;

	if (r->num_chinfo == 0)
		return -1;

	/* find out what changed in latest channel-report */
	for (int i = 0; i < num; i++) {
		struct chan_entry *e = &chinfo[i];

		for (int j = 0; j < r->num_chinfo; j++) {
			if (e->freq != r->chinfo[j].freq)
				continue;

			if (e->dfs && e->dfs_state != r->chinfo[j].dfs_state) {
				dbg("Update dfs state for freq %u\n", e->freq);
				r->chinfo[j].dfs_state = e->dfs_state;
				r->chinfo[j].cac_time = e->cac_time;
				r->chinfo[j].nop_time = e->nop_time;
			}

			if (memcmp(&e->survey, &r->chinfo[j].survey, sizeof(struct wifi_radio_diagnostic)) ||
			    e->noise != r->chinfo[j].noise) {
				dbg("Survey data for freq %u changed.\n", e->freq);
				r->chinfo[j].survey.channel_busy = e->survey.channel_busy;
				r->chinfo[j].survey.tx_airtime = e->survey.tx_airtime;
				r->chinfo[j].survey.rx_airtime = e->survey.rx_airtime;
				r->chinfo[j].survey.cca_time = e->survey.cca_time;
				r->chinfo[j].noise = e->noise;
				r->tsp[j] = time(NULL);
				changed = true;
				break;
			}
		}
	}

	return changed ? 0 : -1;
}

int csmngr_radio_collect_channels_info(struct radio_interface *r)
{
	struct chan_entry chinfo[64] = {0};
	int num_chinfo = ARRAY_SIZE(chinfo);
	int ret = 0;

	if (r->multiband)
		ret = wifi_channels_info_band(r->phyname, r->band, chinfo, &num_chinfo);
	else
		ret = wifi_channels_info(r->phyname, chinfo, &num_chinfo);

	if (ret) {
		warn("%s: %s: failed to collect channels-info\n", __func__, r->ifname);
		return -1;
	}

	csmngr_radio_update_channels_info(r, num_chinfo, chinfo);

#if 1	//debug
	csmngr_dump_channels_info(1, r, r->num_chinfo, r->chinfo, r->scan_channel);
#endif
	return 0;
}

int csmngr_radio_collect_channels_info_t0(struct radio_interface *r)
{
	int ret = 0;

	r->num_chinfo = ARRAY_SIZE(r->chinfo);
	if (r->multiband)
		ret = wifi_channels_info_band(r->phyname, r->band, r->chinfo, &r->num_chinfo);
	else
		ret = wifi_channels_info(r->phyname, r->chinfo, &r->num_chinfo);

	if (ret) {
		r->num_chinfo = 0;
		memset(r->chinfo, 0, sizeof(r->chinfo));
		warn("%s: %s: failed to collect channels-info\n", __func__, r->ifname);
		return -1;
	}

	for (int i = 0; i < r->num_chinfo; i++)
		r->tsp[i] = time(NULL);

	dbg("%s: %s - start-time chinfo data\n", __func__, r->phyname);
#if 1	//debug
	csmngr_dump_channels_info(1, r, r->num_chinfo, r->chinfo, r->scan_channel);
#endif
	return 0;
}

struct channel_data *radio_lookup_channel_data(struct radio_interface *r, uint32_t channel)
{
	for (int i = 0; i < r->num_supp_channels; i++) {
		if (r->chdata[i].channel == channel)
			return &r->chdata[i];
	}

	return NULL;
}

int csmngr_radio_collect_scanresults(struct radio_interface *r)
{
	struct wifi_bss scanres[128] = {0};
	int num = ARRAY_SIZE(scanres);
	int ret = 0;

	ret = wifi_get_scan_results(r->phyname, scanres, &num);
	if (ret) {
		warn("%s: %s: failed to collect scanresults\n", __func__, r->ifname);
		return -1;
	}

	/* update scanresults for the scannned channel */
	struct channel_data *chdata = radio_lookup_channel_data(r, r->scan_channel);
	if (!chdata) {
		warn("%s: %s doesn't have channel-data for %u\n", __func__, r->ifname, r->scan_channel);
		return -1;
	}

	memset(chdata->scanresults, 0, sizeof(chdata->scanresults));
	chdata->num_scanresults = 0;
	for (int i = 0; i < num; i++) {
		if (scanres[i].channel == r->scan_channel) {
			dbg("Scanresult [%d]: bssid = " MACFMT ", channel = %u (bw = %d)\n",
				i, MAC2STR(scanres[i].bssid), scanres[i].channel, scanres[i].curr_bw);

			memcpy(&chdata->scanresults[chdata->num_scanresults++], &scanres[i], sizeof(struct wifi_bss));
		}
	}

	return 0;
}

void csmngr_radio_collect(atimer_t *t)
{
	struct radio_interface *r = container_of(t, struct radio_interface, collecttimer);

	csmngr_radio_collect_channels_info(r);
	//timer_set(t, r->collect_int * 1000);
}

struct radio_interface *csmngr_lookup_radio_by_interface(struct csmngr_private *priv, const char *ifname)
{
	struct radio_interface *r = NULL;

	list_for_each_entry(r, &priv->radiolist, list) {
		for (int i = 0; i < r->info.num_iface; i++) {
			if (!strncmp(r->info.iface[i].name, ifname, 16))
				return r;
		}
	}

	return NULL;
}

struct radio_interface *csmngr_lookup_radio(struct csmngr_private *priv, const char *radio)
{
	struct radio_interface *r = NULL;

	list_for_each_entry(r, &priv->radiolist, list) {
		if (!strncmp(r->ifname, radio, 16))
			return r;
	}

	return NULL;
}

struct radio_interface *csmngr_lookup_radio_by_band(struct csmngr_private *priv, uint8_t band)
{
	struct radio_interface *r = NULL;

	list_for_each_entry(r, &priv->radiolist, list) {
		if ((r->band == BAND_2 && band == 2) ||
		    (r->band == BAND_5 && band == 5) ||
		    (r->band == BAND_6 && band == 6))
			return r;
	}

	return NULL;
}

struct radio_interface *csmngr_alloc_radio(struct csmngr_private *priv, const char *name)
{
	struct radio_interface *r = NULL;
	int br_ifindex;

	if (if_isbridge(name))
		return NULL;

#if 0	//TODO
	if (!is_wifi_radio(name))
		return NULL;
#endif

	r = calloc(1, sizeof(*r));
	if (!r) {
		return NULL;
	}

	INIT_LIST_HEAD(&r->surveylist);
	r->priv = priv;
	snprintf(r->ifname, 16, "%s", name);
	r->ifindex = if_nametoindex(name);
	dbg("%s: %s (%d)\n", __func__, r->ifname, r->ifindex);

	r->refresh_int = CSMNGR_DEFAULT_REFRESH_INT;
	r->collect_int = CSMNGR_DEFAULT_COLLECT_INT;


	if_gethwaddr(name, r->macaddr);
	if_getflags(name, &r->ifflags);
	if_getoperstate(name, &r->operstate);
	if_getcarrier(name, &r->carrier);
	br_ifindex = if_isbridge_interface(name);
	if (br_ifindex > 0) {
		r->is_brif = true;
		r->brport = if_brportnum(name);
		r->br_ifindex = br_ifindex;
	}

	timer_init(&r->refreshtimer, csmngr_radio_periodic_refresh);
	timer_init(&r->collecttimer, csmngr_radio_collect);
	timer_init(&r->runtimer, csmngr_radio_runtimer_cb);

	return r;
}

int is_opclass_chgrp_supported(uint32_t *ctrl_channels, int num, uint32_t *supp_channels)
{
	for (int k = 0; k < 32 && ctrl_channels[k] != 0; k++) {
		int s;

		for (s = 0; s < num; s++) {
			if (supp_channels[s] == ctrl_channels[k])
				break;
		}

		if (s == num)
			return 0;
	}

	return 1;
}

void csmngr_free_radio(struct radio_interface *r)
{
	if (r) {
		timer_del(&r->refreshtimer);
		timer_del(&r->collecttimer);
		timer_del(&r->runtimer);
		free(r);
	}
}

void csmngr_free_radios(struct csmngr_private *priv)
{
	struct radio_interface *r = NULL, *tmp;

	list_for_each_entry_safe(r, tmp, &priv->radiolist, list) {
		list_del(&r->list);
		priv->num_radio--;
		csmngr_free_radio(r);
	}
}

int csmngr_init_radio(struct csmngr_private *priv, struct radio_interface *radio)
{
	//struct chan_entry chinfo[32] = {0};
	//int num_chinfo = ARRAY_SIZE(chinfo);
	uint32_t ch = 0;
	enum wifi_bw bw;
	int noise;
	int ret;

	dbg("%s: ifname = %s --->\n", __func__, radio->ifname);

	ret = wifi_radio_get_ifstatus(radio->phyname, &radio->ifflags);
	if (ret || !(radio->ifflags && IFF_UP)) {
		dbg("%s: %s failed. (ret = %d, ifflags = 0x%x)\n", __func__,
		    radio->ifname, ret, radio->ifflags);
	}

	ret = wifi_radio_is_multiband(radio->phyname, &radio->multiband);
	if (ret) {
		err("%s: %s: failed to determine if multi-band capable\n",
		    __func__, radio->ifname);
	}

	dbg("%s: phy = %s, multiband = %d\n", radio->ifname, radio->phyname, radio->multiband);

#if 0
	wifi_get_supp_channels(radio->phyname, radio->supp_channels,
			       &radio->num_supp_channels, "",
			       radio->band, BW20);
#else
	//TODO: cleanup
	{
		struct wifi_opclass supp_opclass[64] = {0};
		size_t num_opclass = ARRAY_SIZE(supp_opclass);
		uint32_t supp_channels[64] = {0};
		int num_supp_channels = 64;
		uint8_t channels[256] = {0};


		wifi_get_opclass_e4table(radio->band, (1 << BW20), &num_opclass, supp_opclass);
		wifi_get_supp_channels(radio->phyname, supp_channels, &num_supp_channels, "", radio->band, BW20);

		for (int i = 0; i < num_opclass; i++) {
			for (int j = 0; j < supp_opclass[i].opchannel.num; j++) {
				if (is_opclass_chgrp_supported(supp_opclass[i].opchannel.ch[j].ctrl_channels,
								num_supp_channels, supp_channels)) {
					for (int k = 0; k < 32 && supp_opclass[i].opchannel.ch[j].ctrl_channels[k] != 0; k++) {
						channels[supp_opclass[i].opchannel.ch[j].ctrl_channels[k]] = 1;
					}
				}
			}
		}

		dbg("%s: Supported 20MHz channels =\n", radio->ifname);
		for (int i = 0; i < sizeof(channels); i++) {
			if (channels[i] == 1) {
				radio->chdata[radio->num_supp_channels].channel = i;
				//TODO: subsume supp_channels in chd
				radio->supp_channels[radio->num_supp_channels++] = i;
				fprintf(stderr, " %d,", i);
			}
		}
		fprintf(stderr, "\n");
	}
#endif

#if 0	//debug
	dbg("%s: Supported 20MHz channels =\n", radio->ifname);
	for (int i = 0; i < radio->num_supp_channels; i++)
		fprintf(stderr, " %d,", radio->supp_channels[i]);
	fprintf(stderr, "\n");
#endif

	if (radio->multiband) {
		struct wifi_opclass supp_opclass[64] = {0};
		int num_supp_opclass = ARRAY_SIZE(supp_opclass);
		bool has_ap_up = false;
		bool has_sta_up = false;

		ret = wifi_radio_info_band(radio->phyname, radio->band, &radio->info);
		if (ret)
			goto out;

		for (int i = 0; i < radio->info.num_iface; i++) {
			uint32_t ifflags = 0;
			bool up = false;

			ret = wifi_radio_get_ifstatus(radio->info.iface[i].name, &ifflags);
			if (ret || !(ifflags & IFF_UP))
				continue;

			up = true;
			if (radio->info.iface[i].mode == WIFI_MODE_AP) {
				if (!up)
					continue;

				dbg("%s: radio %s has AP interface %s which is UP\n",
				    __func__, radio->phyname, radio->info.iface[i].name);
				has_ap_up = true;
			} if (radio->info.iface[i].mode == WIFI_MODE_STA) {
				if (!up)
					continue;

				dbg("%s: radio %s has STA interface %s which is UP\n",
				    __func__, radio->phyname, radio->info.iface[i].name);
				has_sta_up = true;
			}
		}

		if (!has_ap_up) {
			dbg("Radio %s has no AP interfaces. Skip\n", radio->ifname);
			goto out;
		}

		if (!has_sta_up) {
			dbg("Radio %s has no active STA interface. Enable CS\n", radio->ifname);
			radio->mode = CS_MODE_ENABLED;
		}

		ret |= wifi_get_band_channel(radio->phyname, radio->band, &ch, &bw);
		ret |= wifi_get_band_curr_opclass(radio->phyname, radio->band, &radio->curr_opclass);
		ret |= wifi_get_band_supp_opclass(radio->phyname, radio->band, &num_supp_opclass, supp_opclass);
		if (ret) {
			dbg("%s: %s - failed to get supported opclass\n", __func__, radio->ifname);
			goto out;
		}

		dbg("%s: %s - updating supported opclass\n", __func__, radio->ifname);
		radio->num_supp_opclass = num_supp_opclass;
		memcpy(radio->supp_opclass, supp_opclass, num_supp_opclass * sizeof(struct wifi_opclass));
	} else {
		ret = wifi_radio_info(radio->phyname, &radio->info);
		//TODO: check if has_ap_up and !has_sta_up

		ret |= wifi_get_channel(radio->phyname, &ch, &bw);
		ret |= wifi_get_noise(radio->phyname, &noise);
		ret |= wifi_get_curr_opclass(radio->phyname, &radio->curr_opclass);
		ret |= wifi_get_supp_opclass(radio->phyname, &radio->num_supp_opclass, radio->supp_opclass);
	}

	ret |= wifi_get_band_supp_bandwidths(radio->phyname, radio->band, &radio->supp_bandwidths);

	/* get start-time channels info */
	csmngr_radio_collect_channels_info_t0(radio);

	radio->state = CS_STATE_IDLE;
	radio->scantimeout = SCAN_FINISHED_TIMEOUT;

out:
	dbg("%s: ret = %d\n", __func__, ret);
	return ret;
}

static int csmngr_init_radios(struct csmngr_private *priv)
{
	struct radio_interface *r = NULL;
	int ret = 0;

	if (priv->num_radio == 0 || list_empty(&priv->radiolist))
		return 0;

	list_for_each_entry(r, &priv->radiolist, list) {
		ret |= csmngr_init_radio(priv, r);
	}

	return ret;
}

void csmngr_print_interfaces(struct csmngr_private *priv)
{
	struct radio_interface *e;
	int i = 0;

	if (priv->num_radio == 0 || list_empty(&priv->radiolist))
		return;

	list_for_each_entry(e, &priv->radiolist, list) {
		dbg("%d: ifname = %16s, mac = " MACFMT ", ifindex = %3d, flags = 0x%08x, opstate = 0x%02x, brport = %d\n",
			i++, e->ifname, MAC2STR(e->macaddr), e->ifindex,
			e->ifflags, e->operstate, e->brport);
	}
}

char *csmngr_brport_to_ifname(struct csmngr_private *priv, uint16_t brport)
{
	struct radio_interface *iface;

	if (list_empty(&priv->radiolist))
		return NULL;

	list_for_each_entry(iface, &priv->radiolist, list) {
		if (iface->is_brif && iface->brport == brport)
			return iface->ifname;
	}

	return NULL;
}

int csmngr_handle_ethport_carrier_off(struct csmngr_private *priv,
					const char *ifname)
{
	if (!priv || !ifname)
		return -1;

	//TODO
	return 0;
}

int csmngr_handle_ethport_carrier_on(struct csmngr_private *priv,
				       const char *ifname)
{
	if (!priv || !ifname)
		return -1;

	//TODO
	return 0;
}

void heartbeat_timer_cb(atimer_t *t)
{
	struct csmngr_private *p = container_of(t, struct csmngr_private, hbtimer);

	UNUSED(p);
	switch (signal_pending) {
	case SIGUSR2:
		signal_pending = 0;
		err("%s", "Received SIGUSR2\n");
		break;
	default:
		break;
	}

	timer_set(t, 1000);
}

int csmngr_init_private(struct csmngr_private *priv, void *useropts)
{
	int ret;

	ret = csmngr_init_radios(priv);
	if (ret) {
		warn("%s: failed to properly initialize.\n", __func__);
	}

	return ret;
}

void csmngr_exit_private(struct csmngr_private *priv)
{
	if (priv && priv->num_radio > 0)
		csmngr_free_radios(priv);
}

int csmngr_init(void **priv, void *opts)
{
	struct csmngr_useropts *uopts = (struct csmngr_useropts *)opts;
	struct csmngr_private *p;
	int ret;


	set_sighandler(SIGUSR2, csmngr_sighandler);
	set_sighandler(SIGPIPE, SIG_IGN);
	/* and SIGINT/SIGTERM handlers from uloop cancel uloop */

	*priv = NULL;
	p = calloc(1, sizeof(struct csmngr_private));
	if (!p)
		return -1;

	INIT_LIST_HEAD(&p->radiolist);

	uloop_init();
	p->bus = ubus_connect(uopts->ubus_sockpath);
	if (!p->bus) {
		err("Failed to connect to ubus\n");
		goto out_err;
	}
	ubus_add_uloop(p->bus);

	csmngr_config_defaults(p);
	ret = csmngr_reconfig(p, uopts->confpath, uopts->conffile);
	if (ret) {
		err("Invalid config\n");
		goto out_err;
	}
	csmngr_dump_config(p);

	ret = csmngr_init_private(p, opts);
	if (ret)
		goto out_err;

	ret = csmngr_publish_object(p, uopts->objname);
	if (ret)
		goto out_err;

	ret = csmngr_register_local_events(p);
	if (ret)
		goto out_err;

#if 0	//TODO
	ret = csmngr_register_nlevents(p);
	if (ret)
		goto out_err;
#endif

	/* TODO
	ret = csmngr_start_cmdserver(p, uopts);
	if (ret)
		warn("Failed to start cmdserver.. cli will not be available\n");
	*/

	timer_init(&p->hbtimer, heartbeat_timer_cb);
	timer_set(&p->hbtimer, 1000);

	ret = csmngr_start_engine(p);
	if (ret) {
		warn("Failed to start csmngr.. aborting!\n");
		csmngr_free_radios(p);
		goto out_err;
	}

	*priv = p;
	return 0;

out_err:
	uloop_done();
	free(p);
	return -1;
}

void csmngr_run(void *handle)
{
	UNUSED(handle);

	uloop_run();
}

int csmngr_exit(void *handle)
{
	struct csmngr_private *priv = (struct csmngr_private *)handle;

	if (!priv)
		return 0;

	timer_del(&priv->hbtimer);

	csmngr_unregister_local_events(priv);
#if 0	//TODO
	csmngr_unregister_nlevents(priv);
#endif

	csmngr_remove_object(priv);
	csmngr_exit_private(priv);
	//csmngr_config_free(&priv->cfg);

	ubus_free(priv->bus);
	uloop_done();
	free(priv);
	dbg("csmngr_exit\n");
	return 0;
}

int csmngr_main(void *useropts)
{
	struct csmngr_useropts *opts = (struct csmngr_useropts *)useropts;
	void *ctx;
	int ret;


	if (opts->daemonize)
		do_daemonize(opts->pidfile);

	start_logging(opts);

	ret = csmngr_init(&ctx, opts);
	if (ret) {
		err("%s : Failed to init.\n", CSMNGR_OBJECT);
		return -1;
	}

	csmngr_run(ctx);
	csmngr_exit(ctx);

	stop_logging();

	return 0;
}
