/*
 * ubus.c - ubus interface for status and management.
 *
 * Copyright (C) 2024 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if_arp.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <netlink/route/neighbour.h>

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>

#include <easy/easy.h>

#include "wifi_api.h"
#include "util.h"
#include "debug.h"
#include "timer.h"
#include "csmngr.h"


uint32_t lookup_object(void *bus, const char *objname)
{
	uint32_t id;
	int status;

	status = ubus_lookup_id(bus, objname, &id);
	if (status != UBUS_STATUS_OK) {
		dbg("object '%s' not present!\n", objname);
		return OBJECT_INVALID;
	}

	return id;
}

enum getset_attrs {
	GETSET_ATTR_NAME,
	NUM_ATTRS_GETSET
};

int csmngr_ubus_status(struct ubus_context *ctx, struct ubus_object *obj,
		       struct ubus_request_data *req, const char *method,
		       struct blob_attr *msg)
{
	struct csmngr_private *p = container_of(obj, struct csmngr_private, obj);
	struct blob_buf bb = {0};
	int ret = UBUS_STATUS_OK;

	blob_buf_init(&bb, 0);
	ret = csmngr_status(p, &bb);
	if (ret) {
		if (ret == -EINVAL) {
			ret = UBUS_STATUS_INVALID_ARGUMENT;
		}

		goto out;
	}

	ubus_send_reply(ctx, req, bb.head);
out:
	blob_buf_free(&bb);
	return ret;
}

int csmngr_ubus_results(struct ubus_context *ctx, struct ubus_object *obj,
			struct ubus_request_data *req, const char *method,
			struct blob_attr *msg)
{
	struct csmngr_private *p = container_of(obj, struct csmngr_private, obj);
	struct blob_buf bb = {0};
	int ret = UBUS_STATUS_OK;

	blob_buf_init(&bb, 0);
	ret = csmngr_results(p, &bb);
	if (ret) {
		if (ret == -EINVAL) {
			ret = UBUS_STATUS_INVALID_ARGUMENT;
		}

		goto out;
	}

	ubus_send_reply(ctx, req, bb.head);
out:
	blob_buf_free(&bb);
	return ret;
}


/* scan radio policy */
enum scan_attr {
	SCAN_ATTR_RADIO_NAME,		/* radio name */
	SCAN_ATTR_BAND,			/* band; 2, 5 or 6 (ignored when radio provided) */
	SCAN_ATTR_FREQ,			/* Frequency in MHz. */
	SCAN_ATTR_CHANNEL,		/* 20MHz-channel number (ignored when freq passed)*/
	SCAN_ATTR_DURATION,		/* channel dwell-time in msecs */
	NUM_ATTRS_SCAN,
};

static const struct blobmsg_policy scan_policy[NUM_ATTRS_SCAN] = {
	[SCAN_ATTR_RADIO_NAME] = { .name = "radio", .type = BLOBMSG_TYPE_STRING },
	[SCAN_ATTR_BAND] = { .name = "band", .type = BLOBMSG_TYPE_INT32 },
	[SCAN_ATTR_CHANNEL] = { .name = "channel", .type = BLOBMSG_TYPE_INT32 },
	[SCAN_ATTR_FREQ] = { .name = "freq", .type = BLOBMSG_TYPE_INT32 },
	[SCAN_ATTR_DURATION] = { .name = "duration", .type = BLOBMSG_TYPE_INT32 },
};

int csmngr_ubus_scan(struct ubus_context *ctx, struct ubus_object *obj,
		     struct ubus_request_data *req, const char *method,
		     struct blob_attr *msg)
{
	struct csmngr_private *p = container_of(obj, struct csmngr_private, obj);
	struct blob_buf bb = {0};
	int ret = UBUS_STATUS_OK;

	blob_buf_init(&bb, 0);
	ret = csmngr_scan(p, msg, &bb);
	if (ret) {
		if (ret == -EINVAL) {
			ret = UBUS_STATUS_INVALID_ARGUMENT;
		}

		goto out;
	}

	ubus_send_reply(ctx, req, bb.head);
out:
	blob_buf_free(&bb);
	return ret;
}

/* refresh policy */
enum refresh_attr {
	REFRESH_ATTR_RADIO_NAME,	/* radio name */
	REFRESH_ATTR_BAND,		/* band; 2, 5 or 6 (ignored when radio provided) */
	NUM_ATTRS_REFRESH,
};

static const struct blobmsg_policy refresh_policy[NUM_ATTRS_REFRESH] = {
	[REFRESH_ATTR_RADIO_NAME] = { .name = "radio", .type = BLOBMSG_TYPE_STRING },
	[REFRESH_ATTR_BAND] = { .name = "band", .type = BLOBMSG_TYPE_INT32 },
};

int csmngr_ubus_refresh(struct ubus_context *ctx, struct ubus_object *obj,
			struct ubus_request_data *req, const char *method,
			struct blob_attr *msg)
{
	struct csmngr_private *p = container_of(obj, struct csmngr_private, obj);
	struct blob_buf bb = {0};
	int ret = UBUS_STATUS_OK;

	blob_buf_init(&bb, 0);
	ret = csmngr_refresh(p, msg, &bb);
	if (ret) {
		if (ret == -EINVAL) {
			ret = UBUS_STATUS_INVALID_ARGUMENT;
		}

		goto out;
	}

	ubus_send_reply(ctx, req, bb.head);
out:
	blob_buf_free(&bb);
	return ret;
}

/* dump survey policy */
enum dump_survey_attr {
	DUMP_SURVEY_ATTR_RADIO_NAME,		/* radio name */
	DUMP_SURVEY_ATTR_BAND,			/* band; 2, 5 or 6 (ignored when radio provided) */
	DUMP_SURVEY_ATTR_FREQ,			/* Frequency in MHz. */
	DUMP_SURVEY_ATTR_CHANNEL,		/* 20MHz-channel number (ignored when freq passed) */
	NUM_ATTRS_DUMP_SURVEY,
};

static const struct blobmsg_policy dump_survey_policy[NUM_ATTRS_DUMP_SURVEY] = {
	[DUMP_SURVEY_ATTR_RADIO_NAME] = { .name = "radio", .type = BLOBMSG_TYPE_STRING },
	[DUMP_SURVEY_ATTR_BAND] = { .name = "band", .type = BLOBMSG_TYPE_INT32 },
	[DUMP_SURVEY_ATTR_FREQ] = { .name = "freq", .type = BLOBMSG_TYPE_INT32 },
	[DUMP_SURVEY_ATTR_CHANNEL] = { .name = "channel", .type = BLOBMSG_TYPE_INT32 },
};

int csmngr_ubus_dump_survey(struct ubus_context *ctx, struct ubus_object *obj,
			    struct ubus_request_data *req, const char *method,
			    struct blob_attr *msg)
{
	struct csmngr_private *p = container_of(obj, struct csmngr_private, obj);
	struct blob_buf bb = {0};
	int ret = UBUS_STATUS_OK;

	blob_buf_init(&bb, 0);
	ret = csmngr_dump_survey(p, msg, &bb);
	if (ret) {
		if (ret == -EINVAL) {
			ret = UBUS_STATUS_INVALID_ARGUMENT;
		}

		goto out;
	}

	ubus_send_reply(ctx, req, bb.head);
out:
	blob_buf_free(&bb);
	return ret;
}

int csmngr_publish_object(struct csmngr_private *p, const char *objname)
{
	struct ubus_object_type *obj_type;
	struct ubus_method *obj_methods;
	struct ubus_object *obj;
	struct ubus_method m[] = {
		UBUS_METHOD_NOARG("status", csmngr_ubus_status),
		UBUS_METHOD_NOARG("results", csmngr_ubus_results),
		UBUS_METHOD("refresh", csmngr_ubus_refresh, refresh_policy),
		UBUS_METHOD("scan", csmngr_ubus_scan, scan_policy),
		UBUS_METHOD("dump_survey", csmngr_ubus_dump_survey, dump_survey_policy),
		//UBUS_METHOD("history", csmngr_ubus_history, history_attrs),
		//UBUS_METHOD("set", csmngr_ubus_set, set_attrs),
		//UBUS_METHOD("get", csmngr_ubus_get, get_attrs),
	};
	int num_methods = ARRAY_SIZE(m);
	int ret;

	if (!p->bus) {
		dbg("%s: connect to ubus..\n", __func__);
		p->bus = ubus_connect(NULL);
		if (!p->bus) {
			dbg("Failed to connect to ubus\n");
			return -1;
		}
		ubus_add_uloop(p->bus);
	}

	obj = &p->obj;
	memset(obj, 0, sizeof(*obj));

	obj_type = calloc(1, sizeof(struct ubus_object_type));
	if (!obj_type)
		return -1;

	obj_methods = calloc(num_methods, sizeof(struct ubus_method));
	if (!obj_methods) {
		free(obj_type);
		return -1;
	}

	obj->name = strdup(objname);
	memcpy(obj_methods, m, num_methods * sizeof(struct ubus_method));

	obj->methods = obj_methods;
	obj->n_methods = num_methods;

	obj_type->name = obj->name;
	obj_type->n_methods = obj->n_methods;
	obj_type->methods = obj->methods;
	obj->type = obj_type;
	ret = ubus_add_object(p->bus, obj);
	if (ret) {
		dbg("Failed to add '%s' err = %s\n", objname, ubus_strerror(ret));
		free(obj_methods);
		free(obj_type);

		return ret;
	}

	dbg("Added object '%s'\n", objname);
	return 0;
}

int csmngr_remove_object(struct csmngr_private *p)
{
	if (p->bus) {
		if (p->obj.id != OBJECT_INVALID) {
			ubus_remove_object(p->bus, &p->obj);
			free(p->obj.type);
			free((void *)p->obj.methods);
			free((void *)p->obj.name);
		}
	}

	return 0;
}

static void csmngr_wifi_radio_event_handler(struct csmngr_private *priv,
					    struct blob_attr *msg)
{
	static const struct blobmsg_policy ev_attr[6] = {
		[0] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	struct radio_interface *r = NULL;
	struct blob_attr *tb[3];
	char radio[18] = {0};
	char event[16] = {0};

	blobmsg_parse(ev_attr, 3, tb, blob_data(msg), blob_len(msg));
	if (!tb[0] || !tb[1])
		return;

	strncpy(radio, blobmsg_data(tb[1]), sizeof(radio) - 1);
	strncpy(event, blobmsg_data(tb[0]), sizeof(event) - 1);
	dbg("%s: process '%s' for radio '%s'\n", __func__, event, radio);

	r = csmngr_lookup_radio_by_interface(priv, radio);
	if (!r) {
		dbg("%s: %s not found\n", __func__, radio);
		return;
	}

	if (!strcmp(event, "ap-enabled")) {
		// TODO: if atleast one AP-interface on a radio is enabled,
		// which earlier was not, resume the CS-engine.
		dbg("%s: TODO\n", __func__);

	} else if (!strcmp(event, "ap-disabled")) {
		//TODO: if no more active AP interfaces on a radio is enabled,
		// then pause the CS-engine.
		dbg("%s: TODO\n", __func__);

	} else if (!strcmp(event, "cac-end")) {
		csmngr_process_event_radio_cac_finished(priv, r);

	} else if (!strcmp(event, "radar")) {
		//csmngr_process_event_radio_radar(priv, r);

	} else if (!strcmp(event, "scan_started")) {
		//csmngr_process_event_radio_scan_started(priv, r);

	} else if (!strcmp(event, "scan_finished")) {
		csmngr_process_event_radio_scan_finished(priv, r);
	}
}

static void csmngr_wifi_sta_event_handler(struct csmngr_private *p,
					 struct blob_attr *msg)
{
	static const struct blobmsg_policy ev_attr[3] = {
		[0] = { .name = "ifname", .type = BLOBMSG_TYPE_STRING },
		[1] = { .name = "event", .type = BLOBMSG_TYPE_STRING },
		[2] = { .name = "data", .type = BLOBMSG_TYPE_TABLE },
	};
	bool add = false, del = false;
	struct blob_attr *tb[3];
	char ifname[16] = {0};
	char event[16] = {0};


	blobmsg_parse(ev_attr, 3, tb, blob_data(msg), blob_len(msg));
	if (!tb[0] || !tb[1] || !tb[2])
		return;

	strncpy(ifname,	blobmsg_data(tb[0]), sizeof(ifname) - 1);
	strncpy(event, blobmsg_data(tb[1]), sizeof(event) - 1);

	add = !strcmp(event, "connected");
	del = !strcmp(event, "disconnected");

	if (add || del) {
		struct blob_attr *data[1];
		static const struct blobmsg_policy data_attr[1] = {
			[0] = { .name = "macaddr", .type = BLOBMSG_TYPE_STRING },
		};
		char macstr[18] = {0};
		uint8_t macaddr[6] = {0};

		blobmsg_parse(data_attr, 1, data, blobmsg_data(tb[2]),
			      blobmsg_data_len(tb[2]));

		if (!data[0])
			return;

		strncpy(macstr, blobmsg_data(data[0]), sizeof(macstr) - 1);
		if (!hwaddr_aton(macstr, macaddr))
			return;

		//TODO: account num-stas for radio
	}
}

static void csmngr_local_event_handler(struct ubus_context *ctx,
				     struct ubus_event_handler *e,
				     const char *type, struct blob_attr *msg)
{
	struct csmngr_private *p = container_of(e, struct csmngr_private, evh);
	struct wifi_ev_handler {
		const char *type;
		void (*handler)(struct csmngr_private *, struct blob_attr *);
	} evs[] = {
		{ "wifi.radio", csmngr_wifi_radio_event_handler },
		{ "wifi.sta", csmngr_wifi_sta_event_handler },
	};
	char *str;
	int i;

	str = blobmsg_format_json(msg, true);
	if (!str)
		return;

	//info("Received event: [%s] event = '%s'\n", type, str);
	free(str);

	for (i = 0; i < ARRAY_SIZE(evs); i++) {
		if (!strcmp(type, evs[i].type)) {
			evs[i].handler(p, msg);
			break;
		}
	}
}

int csmngr_register_local_events(struct csmngr_private *p)
{
	int ret;

	if (!p || !p->bus)
		return -1;

	p->evh.cb = csmngr_local_event_handler;
	ret = ubus_register_event_handler(p->bus, &p->evh, "wifi.*");

	return ret;
}

int csmngr_unregister_local_events(struct csmngr_private *p)
{
	if (p)
		ubus_unregister_event_handler(p->bus, &p->evh);

	return 0;
}

int csmngr_send_event(struct csmngr_private *priv, const char *event, void *data)
{
	struct blob_buf b = {0};

	blob_buf_init(&b, 0);
	//TODO
	ubus_send_event(priv->bus, CSMNGR_EVENT, b.head);
	blob_buf_free(&b);

	return 0;
}
